@@ -10,6 +10,7 @@
 #include "btree/update.h"
 
 #include "data/compress.h"
+#include "data/copygc.h"
 #include "data/ec.h"
 #include "data/extents.h"
 #include "data/keylist.h"
@@ -21,6 +22,7 @@
 
 #include "fs/inode.h"
 
+#include "init/dev.h"
 #include "init/error.h"
 #include "init/fs.h"
 
@@ -38,6 +40,13 @@ static const char * const bch2_data_update_type_strs[] = {
 	NULL
 };
 
+static const struct rhashtable_params bch_update_params = {
+	.head_offset		= offsetof(struct data_update, hash),
+	.key_offset		= offsetof(struct data_update, pos),
+	.key_len		= sizeof(struct bpos),
+	.automatic_shrinking	= true,
+};
+
 static void bkey_put_dev_refs(struct bch_fs *c, struct bkey_s_c k, unsigned ptrs_held)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
@@ -95,7 +104,7 @@ static void count_data_update_key_fail(struct data_update *u,
 
 			unsigned ptr_bit = 1;
 			bkey_for_each_ptr_decode(old.k, bch2_bkey_ptrs_c(old), p, entry) {
-				if ((ptr_bit & u->opts.ptrs_rewrite) &&
+				if ((ptr_bit & u->opts.ptrs_kill ) &&
 				    (ptr = bch2_extent_has_ptr(c, old, p, bkey_i_to_s(insert))) &&
 				    !ptr->cached)
 					rewrites_found |= ptr_bit;
@@ -109,17 +118,17 @@ static void count_data_update_key_fail(struct data_update *u,
 
 		bch2_data_update_opts_to_text(&buf, c, &u->op.opts, &u->opts);
 
-		prt_str_indented(&buf, "\nold:    ");
+		prt_str(&buf, "\nold:    ");
 		bch2_bkey_val_to_text(&buf, c, old);
 
-		prt_str_indented(&buf, "\nnew:    ");
+		prt_str(&buf, "\nnew:    ");
 		bch2_bkey_val_to_text(&buf, c, new);
 
-		prt_str_indented(&buf, "\nwrote:  ");
+		prt_str(&buf, "\nwrote:  ");
 		bch2_bkey_val_to_text(&buf, c, wrote);
 
 		if (insert) {
-			prt_str_indented(&buf, "\ninsert: ");
+			prt_str(&buf, "\ninsert: ");
 			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(insert));
 		}
 	}));
@@ -181,15 +190,18 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	 * other updates
 	 * @new: extent with new pointers that we'll be adding to @insert
 	 *
-	 * Fist, drop ptrs_rewrite from @new:
+	 * Fist, drop ptrs_kill from @new:
 	 */
 	const union bch_extent_entry *entry_c;
 	struct extent_ptr_decoded p;
 	struct bch_extent_ptr *ptr;
 
+	if (u->opts.ptrs_kill_ec)
+		bch2_bkey_drop_ec_mask(c, insert, u->opts.ptrs_kill_ec);
+
 	unsigned rewrites_found = 0, ptr_bit = 1;
 	bkey_for_each_ptr_decode(old.k, bch2_bkey_ptrs_c(old), p, entry_c) {
-		if ((ptr_bit & u->opts.ptrs_rewrite) &&
+		if ((ptr_bit & u->opts.ptrs_kill ) &&
 		    (ptr = bch2_extent_has_ptr(c, old, p, bkey_i_to_s(insert)))) {
 			if (ptr_bit & u->opts.ptrs_io_error)
 				bch2_bkey_drop_ptr_noerror(c, bkey_i_to_s(insert), ptr);
@@ -201,23 +213,34 @@ static int data_update_index_update_key(struct btree_trans *trans,
 		ptr_bit <<= 1;
 	}
 
-	if (u->opts.ptrs_rewrite &&
-	    !rewrites_found &&
-	    bch2_bkey_durability(c, k) >= opts.data_replicas) {
-		count_data_update_key_fail(u, k, bkey_i_to_s_c(&new->k_i), insert,
-					   "no rewrites found:");
-		bch2_btree_iter_advance(iter);
-		return 0;
+	if (u->opts.ptrs_kill && !rewrites_found) {
+		int durability = bch2_bkey_durability(trans, k);
+		if (durability < 0)
+			return durability;
+
+		if (durability >= opts.data_replicas) {
+			count_data_update_key_fail(u, k, bkey_i_to_s_c(&new->k_i), insert,
+						   "no rewrites found:");
+			bch2_btree_iter_advance(iter);
+			return 0;
+		}
 	}
 
 	/*
 	 * A replica that we just wrote might conflict with a replica
 	 * that we want to keep, due to racing with another move:
 	 */
-	const struct bch_extent_ptr *ptr_c;
-	bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(&new->k_i), p, entry,
-		((ptr_c = bch2_bkey_has_device_c(c, bkey_i_to_s_c(insert), p.ptr.dev)) &&
-		 !ptr_c->cached));
+	if (!u->opts.no_devs_have) {
+		const struct bch_extent_ptr *ptr_c;
+		bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(&new->k_i), p, entry,
+			((ptr_c = bch2_bkey_has_device_c(c, bkey_i_to_s_c(insert), p.ptr.dev)) &&
+			 !ptr_c->cached));
+	} else {
+		const struct bch_extent_ptr *ptr_c;
+		bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(insert), p, entry,
+			((ptr_c = bch2_bkey_has_device_c(c, bkey_i_to_s_c(&new->k_i), p.ptr.dev)) &&
+			 !ptr_c->cached));
+	}
 
 	if (!bkey_val_u64s(&new->k)) {
 		count_data_update_key_fail(u, k,
@@ -236,7 +259,8 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	extent_for_each_ptr_decode(extent_i_to_s(new), p, entry)
 		bch2_extent_ptr_decoded_append(c, insert, &p);
 
-	bch2_bkey_drop_extra_durability(c, &opts, bkey_i_to_s(insert));
+	try(bch2_bkey_drop_extra_durability(trans, &opts, bkey_i_to_s(insert)));
+
 	bch2_bkey_drop_extra_cached_ptrs(c, &opts, bkey_i_to_s(insert));
 
 	bool should_check_enospc = false;
@@ -297,6 +321,7 @@ static int data_update_index_update_key(struct btree_trans *trans,
 static int __bch2_data_update_index_update(struct btree_trans *trans,
 					   struct bch_write_op *op)
 {
+	struct bch_fs *c = op->c;
 	struct data_update *u = container_of(op, struct data_update, op);
 	int ret = 0;
 
@@ -305,6 +330,16 @@ static int __bch2_data_update_index_update(struct btree_trans *trans,
 				BTREE_ITER_slots|BTREE_ITER_intent);
 
 	while (!bch2_keylist_empty(&op->insert_keys)) {
+		struct bkey_i *insert = bch2_keylist_front(&op->insert_keys);
+
+		if (insert->k.type != KEY_TYPE_extent) {
+			CLASS(bch_log_msg, msg)(c);
+			prt_printf(&msg.m, "Got non-extent key to insert in data update path - confused:\n");
+			bch2_bkey_val_to_text(&msg.m, c, bkey_i_to_s_c(insert));
+			msg.m.suppress = !bch2_count_fsck_err(c, data_update_got_non_extent, &msg.m);
+			return 0;
+		}
+
 		bch2_trans_begin(trans);
 		ret = data_update_index_update_key(trans, u, &iter);
 
@@ -359,19 +394,21 @@ void bch2_data_update_read_done(struct data_update *u)
 		return;
 	}
 
-	if (u->opts.ptrs_io_error) {
+	if (unlikely(u->opts.ptrs_io_error) ){
 		struct bkey_s_c k = bkey_i_to_s_c(u->k.k);
 		struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 		const union bch_extent_entry *entry;
 		struct extent_ptr_decoded p;
 		unsigned ptr_bit = 1;
 
-		guard(rcu)();
 		bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
 			if ((u->opts.ptrs_io_error & ptr_bit) &&
-			    !(u->opts.ptrs_rewrite & ptr_bit)) {
-				u->op.nr_replicas += bch2_extent_ptr_durability(c, &p);
-				u->opts.ptrs_rewrite |= ptr_bit;
+			    !(u->opts.ptrs_kill & ptr_bit)) {
+				int d = bch2_trans_do(c, bch2_extent_ptr_durability(trans, &p));
+
+				if (d >= 0)
+					u->op.nr_replicas += d;
+				u->opts.ptrs_kill |= ptr_bit;
 				bch2_dev_list_drop_dev(&u->op.devs_have, p.ptr.dev);
 			}
 
@@ -390,10 +427,11 @@ void bch2_data_update_read_done(struct data_update *u)
 
 static inline bool should_trace_update_err(struct data_update *u, int ret)
 {
-	if ((u->opts.type == BCH_DATA_UPDATE_reconcile ||
-	     u->opts.type == BCH_DATA_UPDATE_promote) &&
-	    (bch2_err_matches(ret, BCH_ERR_data_update_fail_no_rw_devs) ||
-	     bch2_err_matches(ret, BCH_ERR_insufficient_devices)))
+	if (bch2_err_matches(ret, BCH_ERR_data_update_fail_need_copygc) ||
+	    ((u->opts.type == BCH_DATA_UPDATE_reconcile ||
+	      u->opts.type == BCH_DATA_UPDATE_promote) &&
+	     (bch2_err_matches(ret, BCH_ERR_data_update_fail_no_rw_devs) ||
+	      bch2_err_matches(ret, BCH_ERR_insufficient_devices))))
 		return false;
 
 	return true;
@@ -425,6 +463,12 @@ void bch2_data_update_exit(struct data_update *update, int ret)
 	struct bch_fs *c = update->op.c;
 	struct bkey_s_c k = bkey_i_to_s_c(update->k.k);
 
+	if (update->on_hashtable) {
+		int ret2 = rhltable_remove(&c->update_table, &update->hash, bch_update_params);
+		BUG_ON(ret2);
+		update->on_hashtable = false;
+	}
+
 	if (update->b)
 		atomic_dec(&update->b->count);
 
@@ -482,6 +526,7 @@ int bch2_update_unwritten_extent(struct btree_trans *trans,
 				&update->op.devs_have,
 				update->op.nr_replicas,
 				update->op.nr_replicas,
+				update->op.nr_replicas,
 				update->op.watermark,
 				0, &cl, &wp);
 		if (bch2_err_matches(ret, BCH_ERR_operation_blocked)) {
@@ -544,29 +589,24 @@ void bch2_data_update_opts_to_text(struct printbuf *out, struct bch_fs *c,
 	prt_str(out, bch2_data_update_type_strs[data_opts->type]);
 	prt_newline(out);
 
-	ptr_bits_to_text(out, data_opts->ptrs_rewrite,	"rewrite");
 	ptr_bits_to_text(out, data_opts->ptrs_io_error,	"io error");
 	ptr_bits_to_text(out, data_opts->ptrs_kill,	"kill");
 	ptr_bits_to_text(out, data_opts->ptrs_kill_ec,	"kill ec");
 
-	prt_str_indented(out, "target:\t");
+	prt_str(out, "target:\t");
 	bch2_target_to_text(out, c, data_opts->target);
 	prt_newline(out);
 
-	prt_str_indented(out, "compression:\t");
-	bch2_compression_opt_to_text(out, io_opts->background_compression);
-	prt_newline(out);
-
-	prt_str_indented(out, "opts.replicas:\t");
-	prt_u64(out, io_opts->data_replicas);
-	prt_newline(out);
-
-	prt_str_indented(out, "extra replicas:\t");
+	prt_str(out, "extra replicas:\t");
 	prt_u64(out, data_opts->extra_replicas);
 	prt_newline(out);
 
 	prt_printf(out, "read_dev:\t%i\n", data_opts->read_dev);
 	prt_printf(out, "checksum_paranoia:\t%i\n", data_opts->checksum_paranoia);
+
+	prt_str(out, "io path options:\t");
+	bch2_inode_opts_to_text(out, c, *io_opts);
+	prt_newline(out);
 }
 
 void bch2_data_update_to_text(struct printbuf *out, struct data_update *m)
@@ -574,7 +614,7 @@ void bch2_data_update_to_text(struct printbuf *out, struct data_update *m)
 	bch2_data_update_opts_to_text(out, m->op.c, &m->op.opts, &m->opts);
 	prt_newline(out);
 
-	prt_str_indented(out, "old key:\t");
+	prt_str(out, "old key:\t");
 	bch2_bkey_val_to_text(out, m->op.c, bkey_i_to_s_c(m->k.k));
 	prt_newline(out);
 
@@ -667,128 +707,193 @@ static int bch2_data_update_bios_init(struct data_update *m, struct bch_fs *c,
 	return 0;
 }
 
-struct bch_devs_list bch2_data_update_devs_keeping(struct bch_fs *c,
-						   struct data_update_opts *opts,
-						   struct bkey_s_c k)
-{
-	struct bch_devs_list ret = (struct bch_devs_list) { 0 };
-
-	/* We always rewrite btree nodes entirely */
-	if (bkey_is_btree_ptr(k.k))
-		return ret;
-
-	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
-	unsigned ptr_bit = 1;
-
-	bkey_for_each_ptr(ptrs, ptr) {
-		if (!(ptr_bit & (opts->ptrs_rewrite|
-				 opts->ptrs_kill)))
-			ret.data[ret.nr++] = ptr->dev;
-		ptr_bit <<= 1;
-	}
-
-	return ret;
-}
-
 static unsigned durability_available_on_target(struct bch_fs *c,
 					       enum bch_watermark watermark,
-					       unsigned target)
+					       enum bch_data_type data_type,
+					       unsigned target,
+					       struct bch_devs_list *devs_have,
+					       bool nonblocking,
+					       struct printbuf *trace,
+					       bool *need_copygc)
 {
-	struct bch_devs_mask devs = target_rw_devs(c, BCH_DATA_user, target);
-	unsigned ret = 0;
+	if (trace) {
+		prt_str(trace, "available to write on ");
+		bch2_target_to_text(trace, c, target);
+		prt_newline(trace);
+		printbuf_indent_add(trace, 2);
+		printbuf_atomic_inc(trace);
+	}
+
+	guard(rcu)();
+	struct bch_devs_mask devs = target_rw_devs(c, data_type, target);
+	unsigned durability = 0;
 
 	unsigned i;
 	for_each_set_bit(i, devs.d, BCH_SB_MEMBERS_MAX) {
+		if (bch2_dev_list_has_dev(*devs_have, i))
+			continue;
+
 		struct bch_dev *ca = bch2_dev_rcu_noerror(c, i);
 		if (!ca)
 			continue;
 
-		struct bch_dev_usage usage;
-		bch2_dev_usage_read_fast(ca, &usage);
+		u64 free = nonblocking
+			? dev_buckets_free(ca, watermark)
+			: dev_buckets_available(ca, watermark);
+		if (free)
+			durability += ca->mi.durability;
+		else if (!bch2_copygc_dev_wait_amount(ca)) {
+			*need_copygc = true;
+			bch2_copygc_wakeup(c);
+		}
 
-		if (dev_buckets_free(ca, usage, watermark))
-			ret += ca->mi.durability;
+		if (trace)
+			prt_printf(trace, "%s: %llu\n", ca->name, free);
 	}
 
-	return ret;
+	if (trace) {
+		printbuf_indent_sub(trace, 2);
+		printbuf_atomic_dec(trace);
+	}
+
+	return durability;
 }
 
-static unsigned bch2_bkey_durability_on_target(struct bch_fs *c, struct bkey_s_c k,
+static unsigned bch2_btree_ptr_durability_on_target(struct bch_fs *c, struct bkey_s_c k,
 					       unsigned target)
 {
+	/* Doesn't handle stripe pointers: */
+
 	struct bch_devs_mask devs = target_rw_devs(c, BCH_DATA_user, target);
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 	const union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
 	unsigned durability = 0;
 
-	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
 		if (p.ptr.dev != BCH_SB_MEMBER_INVALID &&
+		    !p.ptr.cached &&
 		    test_bit(p.ptr.dev, devs.d))
-			durability += bch2_extent_ptr_durability(c, &p);
-	}
+			durability += bch2_dev_durability(c, p.ptr.dev);
+
 	return durability;
 }
 
 static int bch2_can_do_write_btree(struct bch_fs *c,
 				   struct bch_inode_opts *opts,
-				   struct data_update_opts *data_opts, struct bkey_s_c k)
+				   struct data_update_opts *data_opts, struct bkey_s_c k,
+				   struct printbuf *trace)
 {
 	enum bch_watermark watermark = data_opts->commit_flags & BCH_WATERMARK_MASK;
+	struct bch_devs_list empty = {};
+	bool need_copygc = false;
 
-	if (durability_available_on_target(c, watermark, data_opts->target) >
-	    bch2_bkey_durability_on_target(c, k, data_opts->target))
+	if (durability_available_on_target(c, watermark, BCH_DATA_btree, data_opts->target, &empty,
+					   data_opts->write_flags & BCH_WRITE_alloc_nowait,
+					   trace, &need_copygc) >
+	    bch2_btree_ptr_durability_on_target(c, k, data_opts->target))
 		return 0;
 
 	if (!(data_opts->write_flags & BCH_WRITE_only_specified_devs)) {
-		unsigned d = bch2_bkey_durability(c, k);
+		unsigned d = bch2_btree_ptr_durability(c, k);
 		if (d < opts->data_replicas &&
-		    d < durability_available_on_target(c, watermark, 0))
+		    d < durability_available_on_target(c, watermark, BCH_DATA_btree, 0, &empty,
+						       data_opts->write_flags & BCH_WRITE_alloc_nowait,
+						       trace, &need_copygc))
 			return 0;
 	}
 
-	return bch_err_throw(c, data_update_fail_no_rw_devs);
+	return __bch2_err_throw(c, !need_copygc
+				? -BCH_ERR_data_update_fail_no_rw_devs
+				: -BCH_ERR_data_update_fail_need_copygc);
 }
 
-int bch2_can_do_write(struct bch_fs *c,
-		      struct bch_inode_opts *opts,
-		      struct data_update_opts *data_opts,
-		      struct bkey_s_c k, struct bch_devs_list *devs_have)
+static int __bch2_can_do_write(struct bch_fs *c,
+			       struct bch_inode_opts *opts,
+			       struct data_update_opts *data_opts,
+			       struct bch_devs_list *devs_have,
+			       struct bkey_s_c k,
+			       struct printbuf *trace)
 {
+	bool btree = bkey_is_btree_ptr(k.k);
 	enum bch_watermark watermark = data_opts->commit_flags & BCH_WATERMARK_MASK;
+	enum bch_data_type data_type = btree
+		? BCH_DATA_btree
+		: BCH_DATA_user;
+	unsigned target = data_opts->write_flags & BCH_WRITE_only_specified_devs
+		? data_opts->target
+		: 0;
 
 	if ((data_opts->write_flags & BCH_WRITE_alloc_nowait) &&
 	    unlikely(c->allocator.open_buckets_nr_free <= bch2_open_buckets_reserved(watermark)))
 		return bch_err_throw(c, data_update_fail_would_block);
 
-	guard(rcu)();
+	if (btree &&
+	    data_opts->type == BCH_DATA_UPDATE_reconcile &&
+	    !bch2_bkey_has_dev_bad_or_evacuating(c, k))
+		return bch2_can_do_write_btree(c, opts, data_opts, k, trace);
 
-	if (bkey_is_btree_ptr(k.k))
-		return bch2_can_do_write_btree(c, opts, data_opts, k);
+	if (trace) {
+		prt_str(trace, "keeping devices: ");
+		bch2_devs_list_to_text(trace, c, devs_have);
+		prt_newline(trace);
 
-	unsigned target = data_opts->write_flags & BCH_WRITE_only_specified_devs
-		? data_opts->target
-		: 0;
-	struct bch_devs_mask devs = target_rw_devs(c, BCH_DATA_user, target);
-
-	darray_for_each(*devs_have, i)
-		if (*i != BCH_SB_MEMBER_INVALID)
-			__clear_bit(*i, devs.d);
-
-	unsigned i;
-	for_each_set_bit(i, devs.d, BCH_SB_MEMBERS_MAX) {
-		struct bch_dev *ca = bch2_dev_rcu_noerror(c, i);
-		if (!ca)
-			continue;
-
-		struct bch_dev_usage usage;
-		bch2_dev_usage_read_fast(ca, &usage);
-
-		if (dev_buckets_free(ca, usage, watermark))
-			return 0;
+		if (data_opts->write_flags & BCH_WRITE_alloc_nowait)
+			prt_str(trace, "nonblocking\n");
 	}
 
-	return bch_err_throw(c, data_update_fail_no_rw_devs);
+	bool need_copygc = false;
+	if (durability_available_on_target(c, watermark, data_type, target, devs_have,
+					   data_opts->write_flags & BCH_WRITE_alloc_nowait,
+					   trace, &need_copygc))
+		return 0;
+
+	return __bch2_err_throw(c, !need_copygc
+				? -BCH_ERR_data_update_fail_no_rw_devs
+				: -BCH_ERR_data_update_fail_need_copygc);
+}
+
+int bch2_can_do_data_update(struct btree_trans *trans,
+			    struct bch_inode_opts *opts,
+			    struct data_update_opts *data_opts,
+			    struct bkey_s_c k,
+			    struct printbuf *trace)
+{
+	struct bch_fs *c = trans->c;
+	struct bch_devs_list devs_have = {};
+	unsigned durability_keeping = 0;
+
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	unsigned ptr_bit = 1;
+
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (p.ptr.dev != BCH_SB_MEMBER_INVALID &&
+		    !(ptr_bit & data_opts->ptrs_kill)) {
+			int d = ptr_bit & data_opts->ptrs_kill_ec
+				? bch2_dev_durability(c, p.ptr.dev)
+				: bch2_extent_ptr_durability(trans, &p);
+			if (d < 0)
+				return d;
+
+			durability_keeping += d;
+			if (!data_opts->no_devs_have)
+				devs_have.data[devs_have.nr++] = p.ptr.dev;
+		}
+
+		ptr_bit <<= 1;
+	}
+
+	if (!bkey_is_btree_ptr(k.k) &&
+	    !data_opts->extra_replicas &&
+	    durability_keeping >= opts->data_replicas)
+		return 0;
+
+	if (trace)
+		prt_printf(trace, "need %u replicas\n", opts->data_replicas - durability_keeping);
+
+	return __bch2_can_do_write(c, opts, data_opts, &devs_have, k, trace);
 }
 
 /*
@@ -822,12 +927,12 @@ static void checksummed_and_non_checksummed_handling(struct data_update *u, stru
 	unsigned ptr_bit = 1;
 
 	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
-		if (ptr_bit & u->opts.ptrs_rewrite) {
+		if (ptr_bit & u->opts.ptrs_kill) {
 			if (!rewrite_found) {
 				rewrite_found = true;
 				u->opts.read_dev = p.ptr.dev;
 			} else {
-				u->opts.ptrs_rewrite &= ~ptr_bit;
+				u->opts.ptrs_kill &= ~ptr_bit;
 			}
 		}
 
@@ -848,6 +953,7 @@ int bch2_data_update_init(struct btree_trans *trans,
 	struct bch_fs *c = trans->c;
 	int ret = 0;
 
+	m->pos = k.k->p;
 	bch2_bkey_buf_init(&m->k);
 	bch2_bkey_buf_reassemble(&m->k, k);
 	k = bkey_i_to_s_c(m->k.k);
@@ -872,7 +978,8 @@ int bch2_data_update_init(struct btree_trans *trans,
 		BCH_WRITE_move|
 		m->opts.write_flags;
 	m->op.compression_opt	= io_opts->background_compression;
-	m->op.watermark		= m->opts.commit_flags & BCH_WATERMARK_MASK;
+	m->op.watermark		= max(m->opts.commit_flags & BCH_WATERMARK_MASK,
+				      BCH_WATERMARK_normal);
 
 	if (k.k->p.snapshot &&
 	    unlikely(ret = bch2_check_key_has_snapshot(trans, iter, k))) {
@@ -886,65 +993,58 @@ int bch2_data_update_init(struct btree_trans *trans,
 		goto out;
 	}
 
-	unsigned durability_have = 0, durability_removing = 0;
+	if (m->opts.extra_replicas) {
+		ret = bch2_disk_reservation_add(c, &m->op.res, k.k->size * m->opts.extra_replicas, 0);
+		if (ret)
+			goto out;
+	}
 
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(bkey_i_to_s_c(m->k.k));
 	const union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
-	unsigned reserve_sectors = k.k->size * data_opts.extra_replicas;
 	unsigned buf_bytes = 0;
 	bool unwritten = false;
+	unsigned durability_keeping = 0;
 
-	if (m->opts.ptrs_rewrite)
+	if (m->opts.ptrs_kill)
 		checksummed_and_non_checksummed_handling(m, ptrs);
 
-	scoped_guard(rcu) {
-		unsigned ptr_bit = 1;
-		bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
-			if (!p.ptr.cached) {
-				if (!(ptr_bit & (m->opts.ptrs_rewrite|
-						 m->opts.ptrs_kill))) {
-					bch2_dev_list_add_dev(&m->op.devs_have, p.ptr.dev);
-					durability_have += bch2_extent_ptr_durability(c, &p);
-				}
-
-				if (ptr_bit & m->opts.ptrs_rewrite) {
-					if (crc_is_compressed(p.crc))
-						reserve_sectors += k.k->size;
-
-					m->op.nr_replicas += bch2_extent_ptr_desired_durability(c, &p);
-					durability_removing += bch2_extent_ptr_desired_durability(c, &p);
-				}
-			} else {
-				if (m->opts.ptrs_rewrite & ptr_bit) {
-					m->opts.ptrs_kill |= ptr_bit;
-					m->opts.ptrs_rewrite ^= ptr_bit;
-				}
+	unsigned ptr_bit = 1;
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (!(ptr_bit & m->opts.ptrs_kill)) {
+			int d = ptr_bit & m->opts.ptrs_kill_ec
+				? bch2_dev_durability(c, p.ptr.dev)
+				: bch2_extent_ptr_durability(trans, &p);
+			if (d < 0) {
+				ret = d;
+				goto out;
 			}
 
-			/*
-			 * op->csum_type is normally initialized from the fs/file's
-			 * current options - but if an extent is encrypted, we require
-			 * that it stays encrypted:
-			 */
-			if (bch2_csum_type_is_encryption(p.crc.csum_type)) {
-				m->op.nonce	= p.crc.nonce + p.crc.offset;
-				m->op.csum_type = p.crc.csum_type;
-			}
-
-			if (p.crc.compression_type == BCH_COMPRESSION_TYPE_incompressible)
-				m->op.incompressible = true;
-
-			buf_bytes = max_t(unsigned, buf_bytes, p.crc.uncompressed_size << 9);
-			unwritten |= p.ptr.unwritten;
-
-			ptr_bit <<= 1;
+			durability_keeping += d;
+			if (!m->opts.no_devs_have)
+				bch2_dev_list_add_dev(&m->op.devs_have, p.ptr.dev);
 		}
+
+		/*
+		 * op->csum_type is normally initialized from the fs/file's
+		 * current options - but if an extent is encrypted, we require
+		 * that it stays encrypted:
+		 */
+		if (bch2_csum_type_is_encryption(p.crc.csum_type)) {
+			m->op.nonce	= p.crc.nonce + p.crc.offset;
+			m->op.csum_type = p.crc.csum_type;
+		}
+
+		if (p.crc.compression_type == BCH_COMPRESSION_TYPE_incompressible)
+			m->op.incompressible = true;
+
+		buf_bytes = max_t(unsigned, buf_bytes, p.crc.uncompressed_size << 9);
+		unwritten |= p.ptr.unwritten;
+
+		ptr_bit <<= 1;
 	}
 
 	if (m->opts.type != BCH_DATA_UPDATE_scrub) {
-		unsigned durability_required = max(0, (int) (io_opts->data_replicas - durability_have));
-
 		/*
 		 * If current extent durability is less than io_opts.data_replicas,
 		 * we're not trying to rereplicate the extent up to data_alloc/replicas.here -
@@ -953,18 +1053,9 @@ int bch2_data_update_init(struct btree_trans *trans,
 		 * Increasing replication is an explicit operation triggered by
 		 * rereplicate, currently, so that users don't get an unexpected -ENOSPC
 		 */
-		m->op.nr_replicas = min(durability_removing, durability_required) +
+		m->op.nr_replicas = max(0, (int) (io_opts->data_replicas - durability_keeping)) +
 			m->opts.extra_replicas;
-
-		/*
-		 * If device(s) were set to durability=0 after data was written to them
-		 * we can end up with a duribilty=0 extent, and the normal algorithm
-		 * that tries not to increase durability doesn't work:
-		 */
-		if (!(durability_have + durability_removing))
-			m->op.nr_replicas = max((unsigned) m->op.nr_replicas, 1);
-
-		m->op.nr_replicas_required = m->op.nr_replicas;
+		m->op.nr_replicas_required = 1;
 
 		/*
 		 * It might turn out that we don't need any new replicas, if the
@@ -972,8 +1063,6 @@ int bch2_data_update_init(struct btree_trans *trans,
 		 * was written:
 		 */
 		if (!m->op.nr_replicas) {
-			m->opts.ptrs_kill |= m->opts.ptrs_rewrite;
-			m->opts.ptrs_rewrite = 0;
 			/* if iter == NULL, it's just a promote */
 			if (iter)
 				ret = bch2_extent_drop_ptrs(trans, iter, k, io_opts, &m->opts);
@@ -996,19 +1085,20 @@ int bch2_data_update_init(struct btree_trans *trans,
 		 *   single durability=2 device)
 		 */
 		if (data_opts.type != BCH_DATA_UPDATE_copygc) {
-			ret = bch2_can_do_write(c, io_opts, &m->opts, k, &m->op.devs_have);
+			ret = __bch2_can_do_write(c, io_opts, &m->opts, &m->op.devs_have, k, NULL);
 			if (ret)
 				goto out;
+
+			guard(rcu)();
+			if (rhltable_lookup(&c->update_table, &m->pos, bch_update_params)) {
+				event_inc(c, data_update_in_flight);
+				ret = bch_err_throw(c, data_update_fail_in_flight);
+				goto out;
+			}
 		}
 
-		if (reserve_sectors) {
-			ret = bch2_disk_reservation_add(c, &m->op.res, reserve_sectors,
-					m->opts.extra_replicas
-					? 0
-					: BCH_DISK_RESERVATION_NOFAIL);
-			if (ret)
-				goto out;
-		}
+		if (!rhltable_insert_key(&c->update_table, &m->pos, &m->hash, bch_update_params))
+			m->on_hashtable = true;
 	} else {
 		if (unwritten) {
 			ret = bch_err_throw(c, data_update_done_unwritten);
@@ -1059,6 +1149,9 @@ int bch2_data_update_init(struct btree_trans *trans,
 	ret = bch2_data_update_bios_init(m, c, io_opts, buf_bytes);
 	if (ret)
 		goto out_nocow_unlock;
+
+	m->rbio.data_update_verify_decompress = m->opts.type == BCH_DATA_UPDATE_scrub;
+
 	return 0;
 out_nocow_unlock:
 	if (c->opts.nocow_enabled)
@@ -1069,6 +1162,12 @@ out:
 	if (!bch2_err_matches(ret, BCH_ERR_transaction_restart))
 		data_update_trace(m, ret);
 
+	if (m->on_hashtable) {
+		int ret2 = rhltable_remove(&c->update_table, &m->hash, bch_update_params);
+		BUG_ON(ret2);
+		m->on_hashtable = false;
+	}
+
 	bkey_put_dev_refs(c, k, m->ptrs_held);
 	m->ptrs_held = 0;
 	bch2_disk_reservation_put(c, &m->op.res);
@@ -1076,3 +1175,17 @@ out:
 
 	return ret;
 }
+
+void bch2_fs_data_update_exit(struct bch_fs *c)
+{
+	if (c->update_table.ht.tbl)
+		rhltable_destroy(&c->update_table);
+}
+
+int bch2_fs_data_update_init(struct bch_fs *c)
+{
+	if (rhltable_init(&c->update_table, &bch_update_params))
+		return bch_err_throw(c, ENOMEM_promote_table_init);
+
+	return 0;
+}
