@@ -6,8 +6,6 @@
  * Copyright 2012 Google, Inc.
  */
 
-#ifndef NO_BCACHEFS_SYSFS
-
 #include "bcachefs.h"
 
 #include "alloc/accounting.h"
@@ -57,6 +55,7 @@
 #include <linux/blkdev.h>
 #include <linux/sort.h>
 #include <linux/sched/clock.h>
+#include <linux/sysfs.h>
 
 
 #define SYSFS_OPS(type)							\
@@ -493,7 +492,7 @@ STORE(bch2_fs)
 		CLASS(bch_log_msg, msg)(c);
 
 		prt_printf(&msg.m, "shutdown by sysfs\n");
-		bch2_fs_emergency_read_only2(c, &msg.m);
+		bch2_fs_emergency_read_only(c, &msg.m);
 	}
 
 #ifdef CONFIG_BCACHEFS_TESTS
@@ -714,11 +713,14 @@ static ssize_t sysfs_opt_store(struct bch_fs *c,
 		BUG();
 	}
 
-	if (!ca)
-		bch2_opt_set_by_id(&c->opts, id, v);
+	if (changed) {
+		if (!ca) {
+			bch2_opt_set_by_id(&c->opts, id, v);
+			clear_bit(id, c->mount_opts.d);
+		}
 
-	if (changed)
 		bch2_opt_hook_post_set(c, ca, 0, id, v);
+	}
 
 	ret = size;
 err:
@@ -934,5 +936,3 @@ struct attribute *bch2_dev_files[] = {
 	&sysfs_write_refs,
 	NULL
 };
-
-#endif  /* _BCACHEFS_SYSFS_H_ */
